/*

Copyright (C) 2000, 2001 Christian Kreibich <kreibich@aciri.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <net/ethernet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <gtk/gtk.h>
#include <callbacks.h>
#include <nd.h>
#include <nd_globals.h>
#include <nd_gui.h>
#include <nd_misc.h>
#include <nd_packet.h>
#include <nd_tracefile.h>
#include <support.h>

#define FDDI_HDR_LEN    13
#define SNAP_LEN        8

struct fddi_header {
  u_char fddi_fc;
  u_char fddi_dhost[6];
  u_char fddi_shost[6];
};

struct snap_header
{
  u_char dsap;
  u_char ssap;
  u_char ctrl;
  u_char snap[5];
};
/*
void        
nd_packet_set_gui(ND_Packet *p)
{
  GtkWidget  *w, *w2;
  int         i;

  if (!p)
    return;

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1");
  
  i = 0;
  w2 = gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), i);
  
  while (w2)
    {
      gtk_widget_hide(w2);
      i++;
      w2 = gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), i);
    }

  nd_packet_set_link_level(p, 1);
}
*/

void 
nd_packet_set_link_level(ND_Packet *p, int upward)
{
  GtkWidget   *w;
  char         s[1024];

  if (!p)
    return;

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1");
  D_ASSERT(w);
  gtk_widget_show(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), FDDI_TAB));


  switch (p->link_prot)
    {
    case ND_PROT_ETHER:
      {
	struct ether_header *eh;
	
	eh = (struct ether_header*)p->link_data;
	
	sprintf(s, "Dest. Addr. (%x:%x:%x:%x:%x:%x)",
		eh->ether_dhost[0]&0xff, eh->ether_dhost[1]&0xff, 
		eh->ether_dhost[2]&0xff, eh->ether_dhost[3]&0xff, 
		eh->ether_dhost[4]&0xff, eh->ether_dhost[5]&0xff);
	
	w = gtk_object_get_data(GTK_OBJECT(toplevel), "link_dest_addr");
	D_ASSERT(w);
	gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
	
	sprintf(s, "Src. Addr. (%x:%x:%x:%x:%x:%x)",
		eh->ether_shost[0]&0xff, eh->ether_shost[1]&0xff, 
		eh->ether_shost[2]&0xff, eh->ether_shost[3]&0xff, 
		eh->ether_shost[4]&0xff, eh->ether_shost[5]&0xff);
	
	w = gtk_object_get_data(GTK_OBJECT(toplevel), "link_source_addr");
	D_ASSERT(w);
	gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
	
	sprintf(s, "Type (%u)", eh->ether_type);

	w = gtk_object_get_data(GTK_OBJECT(toplevel), "link_type");
	D_ASSERT(w);
	gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

      }
      break;
    case ND_PROT_FDDI:
      {
	struct fddi_header  *fh;
	struct snap_header  *sh;
	
	fh = (struct fddi_header*)p->link_data;
	sh = (struct snap_header*)(p->link_data + FDDI_HDR_LEN);
	
      }
      break;
    case ND_PROT_SLIP:
      break;
    case ND_PROT_PPP:
      break;
    default:
    }

  if (upward)
    {
      switch (p->net_prot)
	{
	case ND_PROT_IP:
	  nd_packet_set_ip(p, 1);
	  break;
	default:
	}
    }
  else
    nd_trace_update_tcpdump_packet(p);
}


void 
nd_packet_set_ip(ND_Packet *p, int upward)
{
  GtkWidget   *w;
  char         s[1024];
  struct ip   *iphdr;
  int          opts_len, opts_done, o_len;
  u_char      *opts_p;

  /* If you take the red pill, you'll see that this code is boring
     and painful. If you take the blue one, you'll find yourself
     at the end of this function and will never think of it again,
     hopfully :)
  */

  if (!p)
    return;

  iphdr = nd_packet_ip(p);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1"); 
  D_ASSERT(w);
  gtk_widget_show(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), IP_TAB));
  gtk_notebook_set_page(GTK_NOTEBOOK(w), IP_TAB);
  
  sprintf(s, "IP Version(%u)", iphdr->ip_v);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_v");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  sprintf(s, "Hdr Len (%u)", 4 * iphdr->ip_hl);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_hl");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  if (iphdr->ip_tos & IPTOS_LOWDELAY)
    sprintf(s, "ToS (Min. Delay)");
  else if (iphdr->ip_tos & IPTOS_THROUGHPUT)
    sprintf(s, "ToS (Max. Throughput)");
  else if (iphdr->ip_tos & IPTOS_RELIABILITY)
    sprintf(s, "ToS (Max. Reliability)");
  else if (iphdr->ip_tos & IPTOS_MINCOST)
    sprintf(s, "ToS (Min. Cost)");
  else
    sprintf(s, "ToS (%u)", iphdr->ip_tos);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_tos");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  sprintf(s, "Total Length (%u)", htons(iphdr->ip_len));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_len");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  sprintf(s, "IP ID (%u)", htons(iphdr->ip_id));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_id");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_rf");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_ip_rf_toggled, NULL);
  if (htons(iphdr->ip_off) & IP_RF)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_ip_rf_toggled, NULL);
  
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_df");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_ip_df_toggled, NULL);
  if (htons(iphdr->ip_off) & IP_DF)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_ip_df_toggled, NULL);
  
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_mf");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_ip_mf_toggled, NULL);
  if (htons(iphdr->ip_off) & IP_MF)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_ip_mf_toggled, NULL);
  
  sprintf(s, "Fragment Offset (%u)", (ntohs(iphdr->ip_off) & IP_OFFMASK) * 8);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_off");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  sprintf(s, "Time To Live (%u)", iphdr->ip_ttl);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_ttl");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  switch (iphdr->ip_p)
    {
    case IPPROTO_ICMP:
      sprintf(s, "Protocol (ICMP)");
      break;
    case IPPROTO_IGMP:
      sprintf(s, "Protocol (IGMP)");
      break;
    case IPPROTO_IPIP:
      sprintf(s, "Protocol (IP in IP)");
      break;
    case IPPROTO_TCP:
      sprintf(s, "Protocol (TCP)");
      break;
    case IPPROTO_EGP:
      sprintf(s, "Protocol (EGP)");
      break;
    case IPPROTO_UDP:
      sprintf(s, "Protocol (UDP)");
      break;
    case IPPROTO_IPV6:
      sprintf(s, "Protocol (IPv6)");
      break;
    case IPPROTO_ROUTING:
      sprintf(s, "Protocol (IPv6 Rtg.)");
      break;
    case IPPROTO_FRAGMENT:
      sprintf(s, "Protocol (IPv6 Frg.)");
      break;
    case IPPROTO_RSVP:
      sprintf(s, "Protocol (RSVP)");
      break;
    case IPPROTO_ICMPV6:
      sprintf(s, "Protocol (ICMPv6)");
      break;
    case IPPROTO_NONE:
      sprintf(s, "Protocol (NONE)");
      break;
    case IPPROTO_RAW:
      sprintf(s, "Protocol (RAW)");
      break;
    default:
      sprintf(s, "Protocol (%u)", iphdr->ip_p);
      break;
    }
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_p");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  sprintf(s, "Header Checksum (%u)", htons(iphdr->ip_sum));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_sum");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  if ((iphdr->ip_sum != nd_misc_ip_checksum(p)) &&
      p->is_complete)
    nd_misc_widget_set_red(w, TRUE);
  else
    nd_misc_widget_set_red(w, FALSE);
  
  sprintf(s, "Source Address (%s)", inet_ntoa(iphdr->ip_src));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_src");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  sprintf(s, "Destination Address (%s)", inet_ntoa(iphdr->ip_dst));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "ip_dst");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  
  nd_gui_ip_table_clear_options();
  
  /* Setup options GUI elements manually: */
  
  opts_len = (iphdr->ip_hl*4) - 20;
  opts_done = 0;
  
  if (opts_len > 0)
    {
      opts_p = p->net_data + 20;
      
      while (opts_done < opts_len)
	{
	  switch(*opts_p)
	    {
	    case 0:
	      /* end of option list - RFC791 */
	      opts_done = opts_len;
	      break;
	    case 1:
	      /* No op */
	      opts_p++;
	      opts_done++;
	      nd_gui_ip_table_add_option("N", "No Option",  1);	    
	      break;
	    case 130:
	    case 133:
	      /*Security - see RFC1108*/
	      /*we sanity check this, but otherwise pass it normally*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("Sec", "Security", o_len);	    
	      break;
	    case 131:
	      /*Loose Source and Record Route - RFC791*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("LS&RR", "Loose source route / Record route", o_len);	    
	      break;
	    case 137:
	      /*Strict Source and Record Route - RFC791*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("SS&RR", "Strict source route / record route", o_len);	    
	      break;
	    case 7:
	      /*Record Route - RFC791*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("RR", "Record Route", o_len);	    
	      break;
	    case 136:
	      /*Stream ID - RFC791*/
	      /*we sanity check this, but otherwise pass it normally*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("SID", "Stream ID", o_len);	    
	      break;
	    case 68:
	      /*Internet timestamp - RFC791*/
	      /*harmless...*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("TS", "Internet Timestamp", o_len);	    
	      break;
	    case 148:
	      /*Router Alert - See RFC2113*/
	      /*we sanity check this, but otherwise pass it normally*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("RA", "Router Alert", o_len);	    
	      break;
	    default:
	      opts_p++;
	      o_len = *opts_p;
	      opts_p += o_len-1;
	      opts_done += o_len;
	      nd_gui_ip_table_add_option("?", "Unknown", o_len);	    
	    }

	  if (opts_done > opts_len)
	    {
	      printf("Warning -- bogus options. ...\n");
	    }
	}
    }      

  /* If requested and this is a first fragment
     (or normal packet), set contents too: */
  if ((ntohs(iphdr->ip_off) & IP_OFFMASK) == 0 &&
      (ntohs(iphdr->ip_off) & IP_MF) == 0      &&
      upward)
    {
      switch (iphdr->ip_p)
	{
	case IPPROTO_ICMP:
	  nd_packet_set_icmp(p);
	  break;
	case IPPROTO_TCP:
	  nd_packet_set_tcp(p, 1);
	  break;
	case IPPROTO_UDP:
	  nd_packet_set_udp(p, 1);
	  break;
	default:
	}
    }
  else
    nd_trace_update_tcpdump_packet(p);
}


void 
nd_packet_set_icmp_error_ip(ND_Packet *p)
{
  GtkWidget      *w;
  char            s[1024];
  struct ip      *iphdr;
  struct udphdr  *udphdr;
  int             opts_len, opts_done, o_len;
  u_char         *opts_p;

  /* If you take the red pill, you'll see that this code is boring
     and painful. If you take the blue one, you'll find yourself
     at the end of this function and will never think of it again,
     hopfully :)
  */

  if (!p)
    return;

  iphdr  = nd_packet_icmp_error_ip(p);
  udphdr = nd_packet_icmp_error_udp(p);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1");
  D_ASSERT(w);
  gtk_widget_hide(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), ICMP_TAB));
  gtk_widget_show(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), ICMP_ERROR_TAB));
  gtk_notebook_set_page(GTK_NOTEBOOK(w), ICMP_ERROR_TAB);

  sprintf(s, "IP Version(%u)", iphdr->ip_v);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_v");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Hdr Len (%u)", 4 * iphdr->ip_hl);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_hl");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  if (iphdr->ip_tos & IPTOS_LOWDELAY)
    sprintf(s, "ToS (Min. Delay)");
  else if (iphdr->ip_tos & IPTOS_THROUGHPUT)
    sprintf(s, "ToS (Max. Throughput)");
  else if (iphdr->ip_tos & IPTOS_RELIABILITY)
    sprintf(s, "ToS (Max. Reliability)");
  else if (iphdr->ip_tos & IPTOS_MINCOST)
    sprintf(s, "ToS (Min. Cost)");
  else
    sprintf(s, "ToS (%u)", iphdr->ip_tos);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_tos");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Total Length (%u)", htons(iphdr->ip_len));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_len");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "IP ID (%u)", htons(iphdr->ip_id));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_id");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_rf");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_ip_rf_toggled,
				   GINT_TO_POINTER(ND_PROT_ICMP));

  if (htons(iphdr->ip_off) & IP_RF)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_ip_rf_toggled,
				     GINT_TO_POINTER(ND_PROT_ICMP));

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_df");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_ip_df_toggled,
				   GINT_TO_POINTER(ND_PROT_ICMP));
  if (htons(iphdr->ip_off) & IP_DF)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_ip_df_toggled,
				     GINT_TO_POINTER(ND_PROT_ICMP));

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_mf");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_ip_mf_toggled,
				   GINT_TO_POINTER(ND_PROT_ICMP));
  if (htons(iphdr->ip_off) & IP_MF)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_ip_mf_toggled,
				     GINT_TO_POINTER(ND_PROT_ICMP));

  sprintf(s, "Fragment Offset (%u)", (ntohs(iphdr->ip_off) & IP_OFFMASK) * 8);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_off");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Time To Live (%u)", iphdr->ip_ttl);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_ttl");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  switch (iphdr->ip_p)
    {
    case IPPROTO_ICMP:
      sprintf(s, "Protocol (ICMP)");
      break;
    case IPPROTO_IGMP:
      sprintf(s, "Protocol (IGMP)");
      break;
    case IPPROTO_IPIP:
      sprintf(s, "Protocol (IP in IP)");
      break;
    case IPPROTO_TCP:
      sprintf(s, "Protocol (TCP)");
      break;
    case IPPROTO_EGP:
      sprintf(s, "Protocol (EGP)");
      break;
    case IPPROTO_UDP:
      sprintf(s, "Protocol (UDP)");
      break;
    case IPPROTO_IPV6:
      sprintf(s, "Protocol (IPv6)");
      break;
    case IPPROTO_ROUTING:
      sprintf(s, "Protocol (IPv6 Rtg.)");
      break;
    case IPPROTO_FRAGMENT:
      sprintf(s, "Protocol (IPv6 Frg.)");
      break;
    case IPPROTO_RSVP:
      sprintf(s, "Protocol (RSVP)");
      break;
    case IPPROTO_ICMPV6:
      sprintf(s, "Protocol (ICMPv6)");
      break;
    case IPPROTO_NONE:
      sprintf(s, "Protocol (NONE)");
      break;
    case IPPROTO_RAW:
      sprintf(s, "Protocol (RAW)");
      break;
    default:
      sprintf(s, "Protocol (%u)", iphdr->ip_p);
      break;
    }
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_p");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Header Checksum (%u)", htons(iphdr->ip_sum));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_sum");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  if ((iphdr->ip_sum != nd_misc_icmp_error_ip_checksum(p)) &&
      p->is_complete)
    nd_misc_widget_set_red(w, TRUE);
  else
    nd_misc_widget_set_red(w, FALSE);

  sprintf(s, "Source Address (%s)", inet_ntoa(iphdr->ip_src));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_src");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Destination Address (%s)", inet_ntoa(iphdr->ip_dst));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_ip_dst");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Source Port (%u)", ntohs(udphdr->uh_sport));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_src_port");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Dest Port (%u)", ntohs(udphdr->uh_dport));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_dst_port");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);  
}


void        
nd_packet_set_icmp(ND_Packet *p)
{
  GtkWidget     *w;
  char           s[1024];
  struct icmp   *icmphdr;

  if (!p)
    return;

  icmphdr = nd_packet_icmp(p);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1");
  D_ASSERT(w);
  gtk_widget_show(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), ICMP_TAB));
  gtk_widget_hide(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), ICMP_ERROR_TAB));
  gtk_notebook_set_page(GTK_NOTEBOOK(w), ICMP_TAB);

  switch (icmphdr->icmp_type)
    {
    case ICMP_ECHOREPLY:
      sprintf(s, "Type (Echo Reply)");
      break;
    case ICMP_REDIRECT:
      sprintf(s, "Type (Redirect)");
      break;
    case ICMP_ECHO:
      sprintf(s, "Type (Echo Request)");
      break;
    case ICMP_UNREACH:
      sprintf(s, "Type (Dest. Unreachable)");
      break;
    case ICMP_SOURCEQUENCH:
      sprintf(s, "Type (Sourcequench)");
      break;
    case ICMP_ROUTERADVERT:
      sprintf(s, "Type (Router Adv.)");
      break;
    case ICMP_ROUTERSOLICIT:
      sprintf(s, "Type (Router Sol.)");
      break;
    case ICMP_TIMXCEED:
      sprintf(s, "Type (Time Exceeded)");
      break;
    case ICMP_PARAMPROB:
      sprintf(s, "Type (Param. Problem)");
      break;
    case ICMP_TSTAMP:
      sprintf(s, "Type (Timestamp Req.)");
      break;
    case ICMP_TSTAMPREPLY:
      sprintf(s, "Type (Timestamp Reply)");
      break;
    case ICMP_IREQ:
      sprintf(s, "Type (Info Req.)");
      break;
    case ICMP_IREQREPLY:
      sprintf(s, "Type (Info Reply)");
      break;
    case ICMP_MASKREQ:
      sprintf(s, "Type (Addr. Mask Req.)");
      break;
    case ICMP_MASKREPLY:
      sprintf(s, "Type (Addr. Mask Reply)");
      break;
    default:
      sprintf(s, "Type (%u)", icmphdr->icmp_type);
    }

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_type");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_type");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  switch (icmphdr->icmp_type)
    {
    case ICMP_UNREACH:

      nd_packet_set_icmp_error_ip(p);

      switch (icmphdr->icmp_code)
	{	  
	case ICMP_UNREACH_NET:
	  sprintf(s, "Code (Net Unr.)");
	  break;
	case ICMP_UNREACH_HOST:
	  sprintf(s, "Code (Host Unr.)");
	  break;
	case ICMP_UNREACH_PROTOCOL:
	  sprintf(s, "Code (Protocol Unr.)");
	  break;
	case ICMP_UNREACH_PORT:
	  sprintf(s, "Code (Port Unr.)");
	  break;
	case ICMP_UNREACH_NEEDFRAG:
	  sprintf(s, "Code (Fragm. with DF)");
	  break;
	case ICMP_UNREACH_SRCFAIL:
	  sprintf(s, "Code (Source Route Failed)");
	  break;
	case ICMP_UNREACH_NET_UNKNOWN:
	  sprintf(s, "Code (Net Unknown)");
	  break;
	case ICMP_UNREACH_HOST_UNKNOWN:
	  sprintf(s, "Code (Host Unknown)");
	  break;
	case ICMP_UNREACH_ISOLATED:
	  sprintf(s, "Code (Source Host Isolated)");
	  break;
	case ICMP_UNREACH_NET_PROHIB:
	  sprintf(s, "Code (Dest. Net Prohib.)");
	  break;
	case ICMP_UNREACH_HOST_PROHIB:
	  sprintf(s, "Code (Dest. Host Prohib.)");
	  break;
	case ICMP_UNREACH_TOSNET:
	  sprintf(s, "Code (Net Unr. for ToS)");
	  break;
	case ICMP_UNREACH_TOSHOST:
	  sprintf(s, "Code (Host Unr. for ToS)");
	  break;
	case ICMP_UNREACH_FILTER_PROHIB:
	  sprintf(s, "Code (Prohib. by Filter)");
	  break;
	case ICMP_UNREACH_HOST_PRECEDENCE:
	  sprintf(s, "Code (Host Prec. Violation)");
	  break;
	case ICMP_UNREACH_PRECEDENCE_CUTOFF:
	  sprintf(s, "Code (Prec. Cutoff)");
	  break;
	default:
	  sprintf(s, "Code (%u)!", icmphdr->icmp_code);
	}
      break;

    case ICMP_REDIRECT:

      nd_packet_set_icmp_error_ip(p);

      switch (icmphdr->icmp_code)
	{	  
	case ICMP_REDIRECT_NET:
	  sprintf(s, "Code (Red. Net)");
	  break;
	case ICMP_REDIRECT_HOST:
	  sprintf(s, "Code (Red. Host)");
	  break;
	case ICMP_REDIRECT_TOSNET:
	  sprintf(s, "Code (Red. Net w/ ToS)");
	  break;
	case ICMP_REDIRECT_TOSHOST:
	  sprintf(s, "Code (Red. Host w/ ToS)");
	  break;
	default:
	  sprintf(s, "Code (%u)!", icmphdr->icmp_code);
	}
      break;

    case ICMP_TIMXCEED:

      nd_packet_set_icmp_error_ip(p);

      switch (icmphdr->icmp_code)
	{	  
	case ICMP_TIMXCEED_INTRANS:
	  sprintf(s, "Code (TTL = 0 in Transit)");
	  break;
	case ICMP_TIMXCEED_REASS:
	  sprintf(s, "Code (TTL = 0 in Reassembly)");
	  break;
	default:
	  sprintf(s, "Code (%u)!", icmphdr->icmp_code);
	}
      break;

    case ICMP_PARAMPROB:

      nd_packet_set_icmp_error_ip(p);

      switch (icmphdr->icmp_code)
	{
	case 0:
	  sprintf(s, "Code (IP Header Bad)");
	  break;
	case ICMP_PARAMPROB_OPTABSENT:
	  sprintf(s, "Code (Option Missing)");
	  break;
	default:
	  sprintf(s, "Code (%u)!", icmphdr->icmp_code);
	}
      break;
      
    default:
      sprintf(s, "Code (%u)", icmphdr->icmp_code);
    }

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_code");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_code");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Checksum (%u)", ntohs(icmphdr->icmp_cksum));

  if ((icmphdr->icmp_cksum != nd_misc_icmp_checksum(p)) &&
      p->is_complete)
    {
      w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_cksum");
      D_ASSERT(w);
      nd_misc_widget_set_red(w, TRUE);
      gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
      w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_cksum");
      D_ASSERT(w);
      nd_misc_widget_set_red(w, TRUE);
      gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
    }
  else
    {
      w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_cksum");
      D_ASSERT(w);
      nd_misc_widget_set_red(w, FALSE);
      gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
      w = gtk_object_get_data(GTK_OBJECT(toplevel), "icmp_error_cksum");
      D_ASSERT(w);
      nd_misc_widget_set_red(w, FALSE);
      gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);
    }

  nd_trace_update_tcpdump_packet(p);
}


void 
nd_packet_set_tcp(ND_Packet *p, int upward)
{
  GtkWidget      *w;
  char            s[1024];
  struct tcphdr  *tcphdr;
  struct servent *srvnt = NULL;
  int             opts_len, opts_done, o_len;
  u_char         *opts_p;

  if (!p)
    return;

  tcphdr = nd_packet_tcp(p);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1");
  D_ASSERT(w);
  gtk_widget_show(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), TCP_TAB));
  gtk_notebook_set_page(GTK_NOTEBOOK(w), TCP_TAB);

  if ( (srvnt = getservbyport(tcphdr->th_sport, "tcp")))
    sprintf(s, "Source Port (%u, %s)", htons(tcphdr->th_sport), srvnt->s_name);
  else
    sprintf(s, "Source Port (%u)", htons(tcphdr->th_sport));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_sport");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  if ( (srvnt = getservbyport(tcphdr->th_dport, "tcp")))
    sprintf(s, "Destination Port (%u, %s)", htons(tcphdr->th_dport), srvnt->s_name);
  else
    sprintf(s, "Destination Port (%u)", htons(tcphdr->th_dport));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_dport");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Seq Number (%u)", (u_int32_t)htonl(tcphdr->th_seq));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_seq");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Ack Number (%u)", (u_int32_t)htonl(tcphdr->th_ack));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_ack");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Hdr Len (%u)", (4 * tcphdr->th_off));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_off");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Reserved (%u)", (tcphdr->th_x2 << 2) + ((tcphdr->th_flags >> 6) & 0x03));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_x2");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_urg_flag");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_th_urg_flag_toggled, NULL);
  if (tcphdr->th_flags & TH_URG)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_th_urg_flag_toggled, NULL);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_ack_flag");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_th_ack_flag_toggled, NULL);
  if (tcphdr->th_flags & TH_ACK)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_th_ack_flag_toggled, NULL);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_psh_flag");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_th_psh_flag_toggled, NULL);
  if (tcphdr->th_flags & TH_PUSH)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_th_psh_flag_toggled, NULL);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_rst_flag");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_th_rst_flag_toggled, NULL);
  if (tcphdr->th_flags & TH_RST)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_th_rst_flag_toggled, NULL);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_syn_flag");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_th_syn_flag_toggled, NULL);
  if (tcphdr->th_flags & TH_SYN)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_th_syn_flag_toggled, NULL);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_fin_flag");
  D_ASSERT(w);
  gtk_signal_handler_block_by_func(GTK_OBJECT(w), on_th_fin_flag_toggled, NULL);
  if (tcphdr->th_flags & TH_FIN)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), FALSE);
  gtk_signal_handler_unblock_by_func(GTK_OBJECT(w), on_th_fin_flag_toggled, NULL);

  sprintf(s, "Window Size (%u)", htons(tcphdr->th_win));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_win");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Checksum (%u)", htons(tcphdr->th_sum));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_sum");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  if ((tcphdr->th_sum != nd_misc_tcp_checksum(p)) &&
      p->is_complete)
    nd_misc_widget_set_red(w, TRUE);
  else
    nd_misc_widget_set_red(w, FALSE);

  sprintf(s, "Urgent Pointer (%u)", htons(tcphdr->th_urp));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "th_urp");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  nd_gui_tcp_table_clear_options();

  /* Setup options GUI elements manually: */
  
  opts_len = (tcphdr->th_off*4) - 20;
  opts_done = 0;

  if (opts_len > 0)
    {
      opts_p = p->transp_data + 20;
      
      while (opts_done < opts_len)
	{
	  switch (*opts_p)
	    {
	    case TCPOPT_EOL:
	      opts_done=opts_len;
	      break;
	    case TCPOPT_NOP:
	      opts_p++;
	      opts_done++;
	      nd_gui_tcp_table_add_option("N", "No Option", 1);
	      break;
	    case TCPOPT_MAXSEG:
	      opts_p++;
	      o_len = *opts_p;
	      opts_p++;

	      if (o_len != TCPOLEN_MAXSEG)
		sprintf(s, "MSS(!): %d", (int)htons(*(u_short*)opts_p)); 
	      else
		sprintf(s, "MSS: %d", (int)htons(*(u_short*)opts_p)); 

	      nd_gui_tcp_table_add_option(s, s, o_len);

	      opts_p += o_len-2;
	      opts_done += o_len;
	      break;
	    case TCPOPT_WINDOW:
	      /*WSCALE is only allowed on SYN segments*/
	      opts_p++;
	      o_len = *opts_p;
	      opts_p++;

	      if (o_len != TCPOLEN_WINDOW)
		sprintf(s, "Wscale(!): %d", (int)htons(*(u_short*)opts_p)); 
	      else
		sprintf(s, "Wscale: %d", (int)htons(*(u_short*)opts_p)); 

	      nd_gui_tcp_table_add_option(s, s, o_len);

	      opts_p += o_len-2;
	      opts_done += o_len;
	      break;
	    case TCPOPT_SACK_PERMITTED:
	      /*SACK_PERMITTED is only allowed on SYN segments*/
	      opts_p++;
	      o_len = *opts_p;

	      if (o_len!=TCPOPT_SACK_PERMITTED)
		nd_gui_tcp_table_add_option("SA", "SAck perm.(!)", o_len);
	      else
		nd_gui_tcp_table_add_option("SA", "SAck perm.", o_len);
	 
	      opts_p+=o_len-1;
	      opts_done+=o_len;
	      break;
	    case TCPOPT_SACK:
	      opts_p++;
	      o_len=*opts_p;
	      nd_gui_tcp_table_add_option("SA", "SAck Option", o_len);
	      opts_p+=o_len-1;
	      opts_done+=o_len;
	      break;
	    case TCPOPT_TIMESTAMP:
	      opts_p++;
	      o_len=*opts_p;
	      nd_gui_tcp_table_add_option("TS", "Timestamp", o_len);
	      opts_p+=o_len-1;
	      opts_done+=o_len;
	      break;
	    case TCPOPT_CC:
	      opts_p++;
	      o_len=*opts_p;
	      nd_gui_tcp_table_add_option("CC", "T/TCP Connection Count", o_len);
	      opts_p+=o_len-1;
	      opts_done+=o_len;
	      break;
	    case TCPOPT_CCNEW:
	      opts_p++;
	      o_len=*opts_p;
	      nd_gui_tcp_table_add_option("CN", "T/TCP New Connection Count", o_len);
	      opts_p+=o_len-1;
	      opts_done+=o_len;
	      break;
	    case TCPOPT_CCECHO:
	      opts_p++;
	      o_len=*opts_p;
	      nd_gui_tcp_table_add_option("CC", "T/TCP Connection Count Echo", o_len);
	      opts_p+=o_len-1;
	      opts_done+=o_len;
	      break;
	    case 19:
	      /*MD5 SIGNATURE*/
	      /*we have to pass this or routing protocols may break!*/
	      /*XXXX - we need to not apply any other normalizations if this
		option is present*/
	      opts_p++;
	      o_len=*opts_p;
	      nd_gui_tcp_table_add_option("MD5", "MD5 Signature", o_len);
	      opts_p+=o_len-1;
	      opts_done+=o_len;
	      break;
	    default:
	      o_len=*opts_p;
	      nd_gui_tcp_table_add_option("?", "Unknown", o_len);
	      opts_done+=o_len;
	      break;
	    }
	  
	  if (opts_done>opts_len)
	    {
	      printf("Warning -- bogus TCP options. ...\n");
	    }
	}
    }

  nd_trace_update_tcpdump_packet(p);

  return;
  upward = 0;
}


void 
nd_packet_set_udp(ND_Packet *p, int upward)
{
  GtkWidget      *w;
  char            s[1024];
  struct udphdr  *udphdr;
  struct servent *srvnt = NULL;

  if (!p)
    return;
  
  udphdr = nd_packet_udp(p);

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1");
  D_ASSERT(w);
  gtk_widget_show(gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), UDP_TAB));
  gtk_notebook_set_page(GTK_NOTEBOOK(w), UDP_TAB);

  if ( (srvnt = getservbyport(udphdr->uh_sport, "udp")))
    sprintf(s, "Source Port (%u, %s)", htons(udphdr->uh_sport), srvnt->s_name);
  else
    sprintf(s, "Source Port (%u)", htons(udphdr->uh_sport));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "uh_sport");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  if ( (srvnt = getservbyport(udphdr->uh_dport, "udp")))
    sprintf(s, "Destination Port (%u, %s)", htons(udphdr->uh_dport), srvnt->s_name);
  else
    sprintf(s, "Destination Port (%u)", htons(udphdr->uh_dport));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "uh_dport");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "UDP Length (%u)", htons(udphdr->uh_ulen));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "uh_ulen");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  sprintf(s, "Checksum (%u)", htons(udphdr->uh_sum));
  w = gtk_object_get_data(GTK_OBJECT(toplevel), "uh_sum");
  D_ASSERT(w);
  gtk_label_set_text(GTK_LABEL(GTK_BIN(w)->child), s);

  if ((udphdr->uh_sum != nd_misc_udp_checksum(p)) &&
      p->is_complete)
    nd_misc_widget_set_red(w, TRUE);
  else
    nd_misc_widget_set_red(w, FALSE);

  nd_trace_update_tcpdump_packet(p);

  return;
  upward = 0;
}


void    
nd_packet_set(ND_Packet *p)
{
  GtkWidget  *w, *w2;
  int         i;

  if (!p)
    return;

  w = gtk_object_get_data(GTK_OBJECT(toplevel), "notebook1");
  D_ASSERT(w);
      
  i = 0;
  w2 = gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), i);

  while (w2)
    {
      gtk_widget_hide(w2);
      i++;
      w2 = gtk_notebook_get_nth_page(GTK_NOTEBOOK(w), i);
    }
      
  trace.p = p;
  
  switch (pcap_datalink(trace.pcap))
    {
    case DLT_NULL:
      p->net_data = p->link_data;
      nd_packet_set_ip(p, 1);
      break;
    default:
      nd_packet_set_link_level(p, 1);
    }
}


ND_Packet  *
nd_packet_new(void)
{
  ND_Packet *p;

  p = (ND_Packet*) malloc(sizeof(ND_Packet));
  bzero(p, sizeof(ND_Packet));
  return (p);
}


void        
nd_packet_free(ND_Packet *p)
{
  u_char   *data;

  if (!p)
    return;

  data = nd_packet_get_data(p, NULL);

  FREE(data);      
  FREE(p);
}


ND_Packet*
nd_packet_duplicate(ND_Packet *p)
{
  ND_Packet    *copy;
  u_char       *data;

  if (!p)
    return (NULL);

  copy = nd_packet_new();
  copy->ph = p->ph;

  /* Copy the packet data and hook it into the right place */
  
  data = (u_char*) malloc(sizeof(u_char) * p->ph.caplen);
  if (p->link_data)
    {
      memcpy(data, p->link_data, p->ph.caplen);
      copy->link_data = data;
    }
  else if (p->net_data)
    {
      memcpy(data, p->net_data, p->ph.caplen);
      copy->net_data = data;
    }
  else if (p->transp_data)
    {
      memcpy(data, p->transp_data, p->ph.caplen);
      copy->transp_data = data;
    }
  else
    {
      memcpy(data, p->app_data, p->ph.caplen);
      copy->app_data = data;
    }

  /* Now make sure the offsets of the upper layers are correct. */

  if (p->link_data)
    {
      if (p->net_data)
	copy->net_data = copy->link_data + (p->net_data - p->link_data);
    }
  if (p->net_data)
    {
      if (p->transp_data)
	copy->transp_data = copy->net_data + (p->transp_data - p->net_data);
    }
  if (p->transp_data)
    {
      if (p->app_data)
	copy->app_data = copy->transp_data + (p->app_data - p->transp_data);
    }

  copy->link_prot   = p->link_prot;
  copy->net_prot    = p->net_prot;
  copy->transp_prot = p->transp_prot;
  copy->app_prot    = p->app_prot;

  copy->is_hidden   = 0;

  return (copy);
}


void        
nd_packet_init(ND_Packet *p)
{
  if (!p)
    return;

  p->link_prot = p->net_prot = p->transp_prot = p->app_prot = ND_PROT_NULL;

  if (p->ph.caplen == p->ph.len)
    p->is_complete = TRUE;
  
  switch (pcap_datalink(trace.pcap))
    {
    case DLT_EN10MB:
      {
	struct ether_header *eh;

	p->link_prot = ND_PROT_ETHER;
	eh = (struct ether_header*)p->link_data;

	/* Is this an Ethernet-II frame or a Ethernet 802.3 SNAP frame?*/
	/* Check ether_type > 1500 --> Ethernet II */
	if (htons(eh->ether_type) <= ETHERMTU)
	  {
	    struct snap_header  *sh;
	    
	    sh = (struct snap_header*)(p->link_data + ETHER_HDR_LEN);
	    
	    if (sh->ssap != 0xaa)
	      {
		fprintf(stderr, "It's not Ethernet-II and it's not SNAP - what is it?\n");
		exit(1);
	      }
	    
	    /* Check for IP payload: 0x00-00-00-08-00 SNAP field */
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) | (!sh->snap[4])) && (sh->snap[3]==8))
	      {
		/* It's a SNAP header containing an IP datagram */		

		p->net_data = p->link_data + ETHER_HDR_LEN+SNAP_LEN;
		p->net_prot = ND_PROT_IP;
	      }
	  }
	else 
	  {
	    if (htons(eh->ether_type) == ETHERTYPE_IP)
	      {
		/* It's Ethernet-II containing an IP datagram */
		p->net_data = p->link_data + ETHER_HDR_LEN;
		p->net_prot = ND_PROT_IP;
	      }
	  }
      }
      break;
    case DLT_SLIP:
      {
      }
      break;
    case DLT_PPP:
      {
      }
      break;
    case DLT_FDDI:
      {
	struct fddi_header  *fh;
	struct snap_header  *sh;
	
	p->link_prot = ND_PROT_FDDI;
	fh = (struct fddi_header*)p->link_data;
	sh = (struct snap_header*)(p->link_data + FDDI_HDR_LEN);
	
	/*Is this an FDDI 802.2 or an FDDI SNAP frame?*/
	if ((sh->ssap==0xaa)&&(sh->dsap==0xaa))
	  {
	    /*It's a SNAP frame*/
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) |
		 (sh->snap[4]==0)) &&(sh->snap[3]==8))
	      {
		/*It's a SNAP header containing an IP datagram*/
		p->net_data = p->link_data + FDDI_HDR_LEN+SNAP_LEN;
		p->net_prot = ND_PROT_IP;
	      }
	    else
	      {
		fprintf(stderr,"FDDI frame does not contain IP\n");
	      }
	  }
	else
	  {
	    printf("XXX we don't handle FDDI 802.2 frames yet\n");
	  }
      }
      break;
    case DLT_NULL:
      {
	/* We'll assume it's IP for now */
	p->net_data = p->link_data;
	p->link_data = NULL;
	p->link_prot = ND_PROT_NULL;
      }
      break;
    default:
    }
  
  switch (p->net_prot)
    {
    case ND_PROT_IP:
      {
	struct ip* iphdr = (struct ip*)p->net_data;
	
	/* If this is a first fragment (or normal packet), check contents: */
	if ((ntohs(iphdr->ip_off) & IP_OFFMASK) == 0 &&
	    (ntohs(iphdr->ip_off) & IP_MF) == 0)
	  {
	    switch (iphdr->ip_p)
	      {
	      case IPPROTO_ICMP:
		p->transp_data = p->net_data + 4*iphdr->ip_hl;
		p->transp_prot = ND_PROT_ICMP;
		break;
	      case IPPROTO_TCP:
		p->transp_data = p->net_data + 4*iphdr->ip_hl;
		p->transp_prot = ND_PROT_TCP;
		break;
	      case IPPROTO_UDP:
		p->transp_data = p->net_data + 4*iphdr->ip_hl;
		p->transp_prot = ND_PROT_UDP;
	      default:
	      }
	  }
      }
      break;
    default:
    }

  switch (p->transp_prot)
    {
    case ND_PROT_ICMP:
      p->app_data = NULL;
      break;
    case ND_PROT_TCP:
      p->app_data = NULL;
      break;
    case ND_PROT_UDP:
      p->app_data = NULL;
      break;
    default:
    }  
}


ND_Packet  *
nd_packet_sel_next(ND_Packet *p)
{
  if (!p)
    return (NULL);

  if (trace.apply_to_all)
    return (p->next);

  return (p->sel_next);
}


u_char*
nd_packet_get_data(ND_Packet *p, ND_Layer *l)
{
  if (!p)
    return (NULL);

  if (p->link_data)
    {
      if (l) *l = ND_LINK;
      return (p->link_data);
    }
  else if (p->net_data)
    {
      if (l) *l = ND_NET;
      return (p->net_data);
    }
  else if (p->transp_data)
    {
      if (l) *l = ND_TRANSP;
      return (p->transp_data);
    }
  else if (p->app_data)
    {
      if (l) *l = ND_APP;
      return (p->app_data);
    }

  if (l) *l = ND_NULL;
  return (NULL);
}


guint       
nd_packet_hash(gconstpointer ptr)
{
  ND_Packet *p = (ND_Packet *)ptr;

  if (p->transp_prot == ND_PROT_TCP)
    {
      return(((u_int32_t)(nd_packet_tcp(p)->th_sport) ^
	      (u_int32_t)nd_packet_ip(p)->ip_src.s_addr)   ^
	     ((u_int32_t)(nd_packet_tcp(p)->th_dport) ^
	      (u_int32_t)nd_packet_ip(p)->ip_dst.s_addr));
    }
  else
    {
      printf("Hashing non-TCP packet\n");
      abort();
    }
}


gint        
nd_packet_cmp(gconstpointer ptr1, gconstpointer ptr2)
{
  ND_Packet *p1 = (ND_Packet*)ptr1;
  ND_Packet *p2 = (ND_Packet*)ptr2;

  /*
  printf("Cmp: %u %u %u %u -- %u %u %u %u\n",
	 nd_packet_ip(p1)->ip_src.s_addr,
	 nd_packet_ip(p1)->ip_dst.s_addr,
	 nd_packet_tcp(p1)->th_dport,
	 nd_packet_tcp(p1)->th_sport,
	 nd_packet_ip(p2)->ip_src.s_addr,
	 nd_packet_ip(p2)->ip_dst.s_addr,
	 nd_packet_tcp(p2)->th_dport,
	 nd_packet_tcp(p2)->th_sport);
  */
  if (((nd_packet_ip(p1)->ip_src.s_addr == nd_packet_ip(p2)->ip_src.s_addr) &&
       (nd_packet_ip(p1)->ip_dst.s_addr == nd_packet_ip(p2)->ip_dst.s_addr) &&
       (nd_packet_tcp(p1)->th_sport == nd_packet_tcp(p2)->th_sport)         &&
       (nd_packet_tcp(p1)->th_dport == nd_packet_tcp(p2)->th_dport))         ||
      ((nd_packet_ip(p1)->ip_dst.s_addr == nd_packet_ip(p2)->ip_src.s_addr) &&
       (nd_packet_ip(p1)->ip_src.s_addr == nd_packet_ip(p2)->ip_dst.s_addr) &&
       (nd_packet_tcp(p1)->th_dport == nd_packet_tcp(p2)->th_sport)         &&
       (nd_packet_tcp(p1)->th_sport == nd_packet_tcp(p2)->th_dport)))
    return TRUE;
      
  return FALSE;
}


struct ip      *
nd_packet_ip(ND_Packet *p)
{
  return (struct ip*)(p->net_data);
}


struct udphdr  *
nd_packet_udp(ND_Packet *p)
{
  return (struct udphdr*)(p->transp_data);
}


struct tcphdr  *
nd_packet_tcp(ND_Packet *p)
{
  return (struct tcphdr*)(p->transp_data);
}


struct icmp    *
nd_packet_icmp(ND_Packet *p)
{
  return (struct icmp*)(p->transp_data);
}


struct ip      *
nd_packet_icmp_error_ip(ND_Packet *p)
{
  return (struct ip*)(p->transp_data + 8);
}


struct udphdr  *
nd_packet_icmp_error_udp(ND_Packet *p)
{
  return (struct udphdr*)(p->transp_data + 8 + (nd_packet_icmp_error_ip(p)->ip_hl * 4));
}


gboolean    
nd_packet_is_ip(ND_Packet *p)
{
  if (!p)
    return FALSE;

  return (p->net_prot == ND_PROT_IP);
}


gboolean    
nd_packet_is_tcp(ND_Packet *p)
{
  if (!p)
    return FALSE;

  return (p->transp_prot == ND_PROT_TCP);
}


gboolean    
nd_packet_is_udp(ND_Packet *p)
{
  if (!p)
    return FALSE;

  return (p->transp_prot == ND_PROT_UDP);
}


gboolean    
nd_packet_is_icmp(ND_Packet *p)
{
  if (!p)
    return FALSE;

  return (p->transp_prot == ND_PROT_ICMP);
}


gboolean    
nd_packet_is_icmp_error(ND_Packet *p)
{
  if (!p)
    return (FALSE);

  if (p->transp_prot != ND_PROT_ICMP)
    return (FALSE);

  switch (nd_packet_icmp(p)->icmp_type)
    {
    case ICMP_UNREACH:
    case ICMP_REDIRECT:
    case ICMP_SOURCEQUENCH:
    case ICMP_TIMXCEED:
    case ICMP_PARAMPROB:
      return (TRUE);
      break;
    default:
    }
  
  return (FALSE);
}
